@@ -1,5 +1,5 @@
 /*
- * rtpgstcodecs.h - Headers for RTP gstreamer codec setting
+ * rtpgstcodecs.c - Source for RTP gstreamer codec setting
  *
  * Farsight RTP/AVP/SAVP/AVPF Module
  * Copyright (C) 2005,2006 Collabora Ltd.
@@ -26,6 +26,7 @@
 
 #include "rtpgstcodecs.h"
 
+#include <stdlib.h>
 #include <string.h>
 
 #define GST_ELEMENTS_CONF_FILE "gstelements.conf"
@@ -100,6 +101,7 @@ static gboolean hash_find_func (const gchar *key,
                                 CodecInternal *codec_internal,
                                 const FarsightCodec *codec_to_match);
 
+static gint codecs_lists_ref[FARSIGHT_MEDIA_TYPE_LAST+1] = { 0 };
 static GList *list_codecs[FARSIGHT_MEDIA_TYPE_LAST+1] = { NULL };
 static GHashTable *list_codecs_internal[FARSIGHT_MEDIA_TYPE_LAST+1] = { NULL };
 static GKeyFile *elem_config = NULL;
@@ -233,6 +235,10 @@ load_codecs (FarsightMediaType media_type)
   GList *recv_list = NULL;
   GList *send_list = NULL;
 
+  codecs_lists_ref[media_type]++;
+  g_debug ("%s: refcount for codecs_list is %d for media %d", G_STRFUNC,
+      codecs_lists_ref[media_type], media_type);
+
   /* if already computed just return list */
   if (list_codecs[media_type])
   {
@@ -259,6 +265,7 @@ load_codecs (FarsightMediaType media_type)
   else
   {
     g_warning ("Invalid media type given to load_codecs");
+    codecs_lists_ref[media_type]--;
     return NULL;
   }
 
@@ -269,6 +276,7 @@ load_codecs (FarsightMediaType media_type)
   /* if we can't send or recv let's just stop here */
   if (!recv_list && !send_list)
   {
+    codecs_lists_ref[media_type]--;
     return NULL;
   }
 
@@ -540,16 +548,22 @@ codec_cap_list_intersect (GList *list1, GList *list2)
 void
 unload_codecs (FarsightMediaType media_type)
 {
-  if (list_codecs[media_type]) {
-    farsight_codec_list_destroy (list_codecs[media_type]);
-    list_codecs[media_type] = NULL;
-  }
-  if (list_codecs_internal[media_type])
+  codecs_lists_ref[media_type]--;
+  g_debug ("%s: refcount for codecs_list is %d for media %d", G_STRFUNC,
+      codecs_lists_ref[media_type], media_type);
+  if (!codecs_lists_ref[media_type])
   {
-    g_hash_table_destroy (list_codecs_internal[media_type]);
-    list_codecs_internal[media_type] = NULL;
+    if (list_codecs[media_type]) {
+      farsight_codec_list_destroy (list_codecs[media_type]);
+      list_codecs[media_type] = NULL;
+    }
+    if (list_codecs_internal[media_type])
+    {
+      g_hash_table_destroy (list_codecs_internal[media_type]);
+      list_codecs_internal[media_type] = NULL;
+    }
+    pt_count = 96;
   }
-  pt_count = 96;
 }
 
 /**
@@ -1093,7 +1107,7 @@ create_codec_cap_list (GstElementFactory *factory,
         entry = (CodecCap *)found_item->data;
       }
 
-      const gchar *name = gst_plugin_feature_get_name 
+      const gchar *name = gst_plugin_feature_get_name
         (GST_PLUGIN_FEATURE (factory));
 
       /* HACK ALERT this is a quick hack to ignore rtph263pay because of the
@@ -1102,7 +1116,7 @@ create_codec_cap_list (GstElementFactory *factory,
       {
         continue;
       }
- 
+
       if (!entry)
       {
         entry = g_new0 (CodecCap, 1);
@@ -1492,6 +1506,8 @@ set_options_on_element (GstElementFactory *elem_factory, GstElement
       gboolean bool_key_value;
       gint int_key_value;
       gdouble double_key_value;
+      glong long_key_value;
+      gulong ulong_key_value;
       switch (param_spec->value_type)
       {
         case G_TYPE_STRING:
@@ -1529,9 +1545,19 @@ set_options_on_element (GstElementFactory *elem_factory, GstElement
           g_debug ("%s is a double", keys[i]);
           break;
         case G_TYPE_ULONG:
+          str_key_value = g_key_file_get_value (elem_config, element_name, keys[i],
+              NULL);
+          ulong_key_value = strtoul(str_key_value, NULL, 10);
+          g_value_init (&key_value, G_TYPE_ULONG);
+          g_value_set_ulong (&key_value, ulong_key_value);
           g_debug ("%s is a ulong", keys[i]);
           break;
         case G_TYPE_LONG:
+          str_key_value = g_key_file_get_value (elem_config, element_name, keys[i],
+              NULL);
+          long_key_value = strtol(str_key_value, NULL, 10);
+          g_value_init (&key_value, G_TYPE_LONG);
+          g_value_set_long (&key_value, long_key_value);
           g_debug ("%s is a long", keys[i]);
           break;
         case G_TYPE_INT:
