@@ -106,13 +106,15 @@ struct _FarsightRTPStreamPrivate
 
   guint bus_watch;
 
+  GArray *pending_src_ids;
+
   gint recv_codec_id;
   gint send_codec_id;
   
   gboolean start_me;
   gboolean build_send_pipeline;
 
-  GHashTable *depayloader_chains_info;
+  gboolean stopping;
 
   gboolean prepared; /*set if prepare has been called*/
 #ifdef HAVE_JINGLEP2P
@@ -140,6 +142,14 @@ struct _FarsightRTPStreamPrivate
   gboolean use_upnp;
 };
 
+typedef struct _DepayloaderChainInfo DepayloaderChainInfo;
+struct _DepayloaderChainInfo
+{
+  FarsightStream *stream;
+  GstPad *pad;
+  gint pt;
+};
+
 #define FARSIGHT_RTP_STREAM_GET_PRIVATE(o)  \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), FARSIGHT_TYPE_RTP_STREAM, \
                                 FarsightRTPStreamPrivate))
@@ -167,7 +177,7 @@ static void farsight_rtp_stream_payload_type_change (GstElement *element,
                                                      gint pt,
                                                      gpointer user_data);
 static void farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, 
-                                                             gint id);
+                                                             gint id, GstPad *pad);
 static GstElement *farsight_rtp_stream_get_pipeline (FarsightStream *stream);
 static gboolean farsight_rtp_stream_set_pipeline (FarsightStream *stream,
     GstElement *pipeline);
@@ -213,6 +223,7 @@ static void farsight_rtp_stream_network_error (gpointer stream);
 
 static void farsight_rtp_stream_upnp_send_request(FarsightRTPStream *self,
     FarsightTransportInfo *derived_trans);
+
 static void farsight_rtp_stream_upnp_close_ports (FarsightRTPStream *self);
 
 static void farsight_rtp_stream_prepare_transports (FarsightStream *self);
@@ -377,10 +388,14 @@ farsight_rtp_stream_init (FarsightRTPStream *self)
     self->priv->sending = TRUE;
   }
 
+  self->priv->stopping = FALSE;
+
   self->priv->prepared = FALSE;
 
   self->priv->bus_watch = 0;
 
+  self->priv->pending_src_ids = g_array_new (FALSE, FALSE, sizeof (gint));
+
 #ifdef HAVE_JINGLEP2P
   self->priv->socket_client = NULL;
 #endif
@@ -414,7 +429,6 @@ farsight_rtp_stream_init (FarsightRTPStream *self)
     }
   }
 #endif
-
 }
 
 static GObject *
@@ -474,6 +488,20 @@ farsight_rtp_stream_constructor (GType type,
   return obj;
 }
 
+static void
+remove_pending_mainloop_sources (FarsightRTPStreamPrivate *priv)
+{
+  gint i;
+  if (priv->pending_src_ids->len)
+  {
+    for (i = 0; i < priv->pending_src_ids->len; i++)
+    {
+      g_source_remove (g_array_index (priv->pending_src_ids, gint, i));
+    }
+    g_array_remove_range (priv->pending_src_ids, 0, priv->pending_src_ids->len);
+  }
+}
+
 static void
 farsight_rtp_stream_dispose (GObject *object)
 {
@@ -493,6 +521,8 @@ farsight_rtp_stream_dispose (GObject *object)
 
   unload_codecs (media_type);
 
+  remove_pending_mainloop_sources (self->priv);
+
   /* chain up to parent */
   G_OBJECT_CLASS (rtp_stream_parent_class)->dispose (object);
 }
@@ -505,6 +535,9 @@ farsight_rtp_stream_finalize (GObject *object)
   g_return_if_fail (self != NULL);
   g_return_if_fail (FARSIGHT_IS_RTP_STREAM (self));
 
+  if (self->priv->pending_src_ids)
+    g_array_free (self->priv->pending_src_ids, TRUE);
+
   if (self->priv->local_candidates)
     farsight_transport_list_destroy (self->priv->local_candidates);
 
@@ -517,11 +550,6 @@ farsight_rtp_stream_finalize (GObject *object)
   if (self->priv->remote_codecs)
     farsight_codec_list_destroy (self->priv->remote_codecs);
 
-  if (self->priv->depayloader_chains_info)
-  {
-    g_hash_table_destroy (self->priv->depayloader_chains_info);
-  }
-
   if (self->priv->active_native_candidate) {
     g_free (self->priv->active_native_candidate);
   }
@@ -664,6 +692,50 @@ farsight_rtp_stream_bus_watch_cb (GstBus *bus, GstMessage *message,
   return TRUE;
 }
 
+static gboolean
+call_create_recv_pipeline (gpointer user_data)
+{
+  DepayloaderChainInfo *chain_info = (DepayloaderChainInfo *)user_data;
+
+  g_debug ("%s (%d): calling create_recv_pipeline from main thread",
+      __FUNCTION__, __LINE__);
+  farsight_rtp_stream_create_new_pt_recv_pipeline(chain_info->stream,
+      chain_info->pt, chain_info->pad);
+
+  g_free (chain_info);
+
+  return FALSE;
+}
+
+static void
+rtpdemux_pad_blocked_callback (GstPad *pad, gboolean blocked,
+    gpointer user_data)
+{
+  DepayloaderChainInfo *chain_info = (DepayloaderChainInfo *)user_data;
+  FarsightRTPStream *self = FARSIGHT_RTP_STREAM (chain_info->stream);
+
+  if (self->priv->stopping)
+  {
+    g_debug ("%s (%d): stream is stopping, canceling action",
+        __FUNCTION__, __LINE__);
+    g_free (chain_info);
+    return;
+  }
+
+  if (blocked)
+  {
+    gint src_id;
+    g_debug ("%s (%d): rtpdemux pad blocked, idle_adding create_recv_pipeline",
+        __FUNCTION__, __LINE__);
+    src_id = g_idle_add (call_create_recv_pipeline, user_data);
+    g_array_append_val (self->priv->pending_src_ids, src_id);
+  }
+  else
+  {
+    g_free (chain_info);
+  }
+}
+
 static void
 farsight_rtp_stream_new_payload_type (GstElement *element, gint pt, 
                                       GstPad *pad, gpointer user_data)
@@ -686,22 +758,24 @@ farsight_rtp_stream_new_payload_type (GstElement *element, gint pt,
   }
   */
 
-  if (!self->priv->depayloader_chains_info) {
-    self->priv->depayloader_chains_info =
-        g_hash_table_new (g_direct_hash, g_direct_equal);
-  }
   chain_info = g_new0 (DepayloaderChainInfo, 1);
+  chain_info->stream = (FarsightStream *)self;
+  chain_info->pt = pt;
   chain_info->pad = pad;
-  g_hash_table_insert (self->priv->depayloader_chains_info,
-                       (gpointer) pt, (gpointer) chain_info);
+
+  gst_object_ref (GST_OBJECT (pad));
+  g_debug ("%s (%d): received stream with new pt, blocking rtpdemux",
+      __FUNCTION__, __LINE__);
+  gst_pad_set_blocked_async (pad, TRUE, rtpdemux_pad_blocked_callback,
+      chain_info);
 }
 
 static void
 farsight_rtp_stream_payload_type_change (GstElement *element, gint pt,
                                          gpointer user_data)
 {
-  FarsightStream *self = FARSIGHT_STREAM (user_data);
-  farsight_rtp_stream_create_new_pt_recv_pipeline(self, pt);
+  g_debug ("%s (%d): received stream payload changed to %d",
+      __FUNCTION__, __LINE__, pt);
 }
 
 static gboolean
@@ -717,16 +791,14 @@ error_on_create_recv_pipeline (gpointer data)
 }
 
 static void
-farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id)
+farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream,
+    gint id, GstPad *pad)
 {
   FarsightRTPStream *self = FARSIGHT_RTP_STREAM (stream);
-  DepayloaderChainInfo *chain_info = NULL;
   gchar *name = NULL;
   GstElement *codec_bin = NULL;
   CodecInternal *codec_internal = NULL;
 
-  /* WARNING This function gets called in the recv pipeline thread */
-
   codec_internal = lookup_codec_by_pt (farsight_stream_get_media_type (stream),
       id);
   if (!codec_internal) {
@@ -736,8 +808,12 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
   g_debug ("%s (%d): active PT change to %d", __FUNCTION__, __LINE__, id);
 
-  chain_info = g_hash_table_lookup (self->priv->depayloader_chains_info,
-          (gpointer) ((gint) id));
+  if (!self->priv->pipeline)
+  {
+    g_warning ("%s (%d): Pipeline has dissappeared, not doing anything",
+        __FUNCTION__, __LINE__);
+    return;
+  }
 
   /* Let's see if a recv codec bin for this codec already exists */
   name = g_strdup_printf ("recv%d", id);
@@ -763,12 +839,14 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
      * are playing */
     farsight_stream_signal_codec_changed (stream, id);
 
+    /* TODO why do I go to PAUSED here? */
     gst_element_set_state (codec_bin, GST_STATE_PAUSED);
 
-    gchar *padname = gst_pad_get_name (chain_info->pad);
+    gchar *padname = gst_pad_get_name (pad);
     gst_element_link_pads (self->priv->rtpdemux, padname, GST_ELEMENT(codec_bin),
             "sink");
 
+    /* TODO why do I go to PLAYING before linking a sink (if required)? */
     gst_element_set_state (codec_bin, GST_STATE_PLAYING);
 
     g_free (padname);
@@ -814,13 +892,23 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
         gst_object_unref (GST_OBJECT (old_codec_bin));
       }
-      gst_element_set_state (self->priv->sink, GST_STATE_READY);
+      /* TODO do we really need to be READY before linking if the codec_bin is
+       * not PLAYING ? */
+      if (!gst_element_set_state (self->priv->sink, GST_STATE_READY))
+      {
+        g_warning ("%s: Sink state change to READY failed!", G_STRFUNC);
+        goto error;
+      }
       gst_element_link_filtered (GST_ELEMENT (codec_bin), self->priv->sink, self->priv->sink_filter);
-      gst_element_set_state (self->priv->sink, GST_STATE_PLAYING);
+      if (!gst_element_set_state (self->priv->sink, GST_STATE_PLAYING))
+      {
+        g_warning ("%s: Sink state change to PLAYING failed!", G_STRFUNC);
+        goto error;
+      }
 
       gst_object_unref (GST_OBJECT (sinkpad));
     }
-    else 
+    else
     {
       g_warning("received stream while sink unset, blocking recv pipeline");
       GstPad *codec_bin_src_pad =
@@ -833,12 +921,17 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
   /* restart self pipeline */
   //gst_xml_write_file (GST_ELEMENT (self->priv->pipeline), stdout);
-  g_debug ("%s: setting state to playing", G_STRFUNC);
+  g_debug ("%s: unblocking rtpdemux src pad", G_STRFUNC);
+  gst_pad_set_blocked_async (pad, FALSE, blocked_cb, NULL);
   return;
 
 error:
-  g_warning ("%s: PT change failed", G_STRFUNC);
-  g_idle_add (error_on_create_recv_pipeline, stream);
+  {
+    gint src_id;
+    g_warning ("%s: PT change failed", G_STRFUNC);
+    src_id = g_idle_add (error_on_create_recv_pipeline, stream);
+    g_array_append_val (self->priv->pending_src_ids, src_id);
+  }
 }
 
 static GstElement*
@@ -1036,20 +1129,21 @@ farsight_rtp_stream_get_active_codec (FarsightStream *stream)
   return self->priv->send_codec_id; 
 }
 
-static void
-unblocked_cb (GstPad *pad, gboolean blocked,
-    gpointer user_data)
-{
-  g_debug ("Pad unblocked successfully");
-  gst_object_unref (GST_OBJECT (pad));
-}
-
 static void
 blocked_cb (GstPad *pad, gboolean blocked,
     gpointer user_data)
 {
-  g_debug ("Pad blocked successfully");
+  gchar *name = gst_pad_get_name (pad);
+  if (blocked)
+  {
+    g_debug ("Pad %s blocked successfully", name);
+  }
+  else
+  {
+    g_debug ("Pad %s unblocked successfully", name);
+  }
   gst_object_unref (GST_OBJECT (pad));
+  g_free (name);
 }
 
 /* this function will be called when a block has succeeded, it will unlink the
@@ -1071,7 +1165,7 @@ static void
 unlink_and_replace (GstPad *element1_src_pad, gboolean blocked,
     gpointer user_data)
 {
-  g_debug ("Blocked pad successfully, unlinking and replacing downstream");
+  g_debug ("%s: Blocked pad successfully, unlinking and replacing downstream", G_STRFUNC);
   GstPad *element2_sink_pad = NULL;
   element2_sink_pad = gst_pad_get_peer (element1_src_pad);
 
@@ -1114,7 +1208,7 @@ unlink_and_replace (GstPad *element1_src_pad, gboolean blocked,
     //gst_element_sync_state_with_parent (element4);
     gst_element_set_state (element4, GST_STATE_PLAYING);
     gst_object_unref (GST_OBJECT (element1));
-    gst_pad_set_blocked_async (element1_src_pad, FALSE, unblocked_cb, NULL);
+    gst_pad_set_blocked_async (element1_src_pad, FALSE, blocked_cb, NULL);
     gst_object_unref (GST_OBJECT (element4));
   }
 
@@ -1134,6 +1228,17 @@ static void
 unlink_and_replace_sink (GstPad *element1_src_pad, gboolean blocked,
     gpointer user_data)
 {
+  gchar *name = gst_pad_get_name (element1_src_pad);
+  if (blocked)
+  {
+    g_debug ("%s: Pad %s blocked successfully", G_STRFUNC, name);
+  }
+  else
+  {
+    g_debug ("%s: Pad %s unblocked successfully", G_STRFUNC, name);
+  }
+  g_free (name);
+
   replace_sink_data *data = (replace_sink_data *)user_data;
   FarsightRTPStream *self = data->self;
 
@@ -1445,7 +1550,7 @@ static gboolean
 farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
 {
   FarsightRTPStream *self = (FarsightRTPStream*) stream;
-  g_debug ("%s (%d): setting sink", __FUNCTION__, __LINE__);
+  g_debug ("%s (%d): setting sink %s", __FUNCTION__, __LINE__, sink?gst_element_get_name (sink):"NULL");
 
   GstElement *codec_bin = NULL;
   if (self->priv->pipeline)
@@ -1469,6 +1574,7 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
       user_data->self = self;
       user_data->sink = NULL;
 
+      g_debug ("%s: blocking codec_bin_src_pad and removing old sink", G_STRFUNC);
       gst_pad_set_blocked_async (codec_bin_src_pad, TRUE, unlink_and_replace_sink,
           NULL);
       self->priv->sink = NULL;
@@ -1487,7 +1593,8 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
         gst_element_set_state (sink, GST_STATE_READY);
         gst_element_link (codec_bin, sink);
         gst_element_set_state (sink, GST_STATE_PLAYING);
-        gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb,
+        g_debug ("%s: unblocking codec_bin_src_pad and setting new sink", G_STRFUNC);
+        gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, blocked_cb,
           NULL);
         gst_object_ref (sink);
         self->priv->sink = sink;
@@ -1495,7 +1602,7 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
       else
       {
         /* we want to block, remove old sink, add new sink, link and unblock */
-        g_debug ("sink already exists, replacing");
+        g_debug ("%s: sink already present, replacing old one", G_STRFUNC);
         replace_sink_data *user_data = g_new0 (replace_sink_data, 1);
         user_data->self = self;
         user_data->sink = sink;
@@ -1507,7 +1614,7 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
           gst_element_set_state (sink, GST_STATE_READY);
           gst_element_link (codec_bin, sink);
           gst_element_set_state (sink, GST_STATE_PLAYING);
-          gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb,
+          gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, blocked_cb,
               NULL);
           self->priv->sink = sink;
         }
@@ -1517,6 +1624,7 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
   }
   else
   {
+    g_debug ("%s: No codec bin present, setting new sink for future use", G_STRFUNC);
     self->priv->sink = sink;
     if (sink)
     {
@@ -1848,6 +1956,7 @@ farsight_rtp_stream_build_send_pipeline (FarsightRTPStream *self)
       GstPad *ghostpad = gst_element_get_pad (self->priv->pipeline, "sink");
       GstPad *codec_bin_sink_pad = gst_element_get_pad (codec_bin, "sink");
       gst_ghost_pad_set_target ((GstGhostPad *)ghostpad, codec_bin_sink_pad);
+      gst_pad_set_active (ghostpad, TRUE);
       gst_element_link_filtered (self->priv->src, self->priv->pipeline,
           self->priv->src_filter);
       gst_object_unref (ghostpad);
@@ -1912,15 +2021,12 @@ farsight_rtp_stream_start (FarsightStream *stream)
   g_return_val_if_fail (self != NULL, FALSE);
   if (self->priv->pipeline == NULL ||
       self->priv->rtpbin == NULL ||
-      self->priv->send_codec_bin == NULL ||
       farsight_stream_get_state (stream) != FARSIGHT_STREAM_STATE_CONNECTED)
   {
     self->priv->start_me = TRUE;
     return FALSE;
   }
 
-  self->priv->start_me = FALSE;
-
   gst_element_set_state (self->priv->pipeline, GST_STATE_PLAYING);
 
   return TRUE;
@@ -1939,6 +2045,8 @@ farsight_rtp_stream_stop (FarsightStream *stream)
 
   g_return_if_fail (stream != NULL);
 
+  self->priv->stopping = TRUE;
+
   if (self->priv->timeout_src != -1)
   {
       g_source_remove (self->priv->timeout_src);
@@ -1949,6 +2057,9 @@ farsight_rtp_stream_stop (FarsightStream *stream)
   {
     g_debug ("%s (%d): stopping media pipeline", __FUNCTION__, __LINE__);
 
+    /* let's make sure we are not blocking the source */
+    farsight_rtp_stream_set_sending(FARSIGHT_STREAM (self), TRUE);
+
     /* let's unlink from the source first */
     farsight_rtp_stream_unlink_source (self);
 
@@ -2018,11 +2129,6 @@ farsight_rtp_stream_stop (FarsightStream *stream)
     self->priv->pipeline = NULL;
     self->priv->src = NULL;
     self->priv->sink = NULL;
-
-    if (self->priv->depayloader_chains_info) {
-      g_hash_table_destroy (self->priv->depayloader_chains_info);
-      self->priv->depayloader_chains_info = NULL;
-    }
   }
   else
   {
@@ -2038,6 +2144,8 @@ farsight_rtp_stream_stop (FarsightStream *stream)
     }
   }
 
+  remove_pending_mainloop_sources (self->priv);
+
 #ifdef HAVE_JINGLEP2P
   if (self->priv->socket_client)
   {
@@ -2105,7 +2213,7 @@ farsight_rtp_stream_set_sending (FarsightStream *stream, gboolean sending)
   if (sending)
   {
     g_debug ("Setting sending to %d", sending);
-    gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb, NULL);
+    gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, blocked_cb, NULL);
    /* g_object_set (G_OBJECT (stream), "direction",
         FARSIGHT_STREAM_DIRECTION_BOTH, NULL);*/
   }
@@ -2512,6 +2620,8 @@ farsight_rtp_stream_prepare_transports (FarsightStream *stream)
 
   /* We can already create the core pipeline now */
   farsight_rtp_stream_build_base_pipeline (self);
+
+  gst_element_set_state (self->priv->pipeline, GST_STATE_PLAYING);
 }
 
 static gboolean 
